/* -*-C-*- */

/*  kcond.c  */

/*
 * Author: Nikita Danilov <NikitaDanilov@yahoo.COM>
 * Keywords: kernel condition variables
 *
 * Copyright (C) 2000 Namesys (Hans Reiser)
 *
 * This file is part of kcond.
 *
 * Kcond is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */

/* $Id$ */
static char cvsid[] = "@(#)$Id$";
static char debugFileId[] = __FILE__;

#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/malloc.h>
#include <linux/interrupt.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/proc_fs.h>
#include <linux/init.h>

#include <asm/errno.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <asm/byteorder.h>
#include <asm/uaccess.h>
#include <asm/unaligned.h>

#include "kcond.h"


#if defined( KCOND_STATS )
static struct
{
  unsigned long created;
  unsigned long active;
  unsigned long total_waits;
  unsigned long active_waits;
  unsigned long interrupted_waits;
  unsigned long signals;
  unsigned long broadcasts;
  unsigned long lost_signals;
  unsigned long lost_broadcasts;
  unsigned long preventive_wakeups;
  unsigned long pred_swtch_skip;
} kcond_stats;

/* KCOND_STATS */
#endif

MODULE_AUTHOR( "Nikita Danilov NikitaDanilov@Yahoo.COM" );
MODULE_DESCRIPTION( "Condition variables" );

#define SET_DINFO( obj )													\
   DEXP( ( obj ) -> dinfo.return_address = __builtin_return_address( 0 ) ;	\
		 ( obj ) -> dinfo.pid = ( current != NULL ) ? current -> pid : -1 )

static int __kcond_wait( kcond_t *cvar, struct kcond_queue_link *link );
static int __kcond_synchro_lock( kcond_synchro_t synchro, void *obj,
								 kcond_t *cvar );
static int __kcond_synchro_unlock( kcond_synchro_t synchro, void *obj,
								   kcond_t *cvar );

kcond_t *kcond_init( kcond_t *cvar )
{
  ASSERT( cvar != NULL );

  cvar -> lock = SPIN_LOCK_UNLOCKED;
  cvar -> queue = NULL;
  SEXP( kcond_stats.created++ );
  SEXP( kcond_stats.active++ );
  SET_DINFO( cvar );
  return cvar;
}

kcond_t kcond_static_init()
{
  kcond_t result;

  return *kcond_init( &result );
}

int kcond_destroy( kcond_t *cvar )
{
  SEXP( kcond_stats.active-- );
  return kcond_are_waiters( cvar ) ? -EBUSY : 0;
}

int kcond_wait( kcond_t *cvar, struct semaphore *sem )
{
  struct kcond_queue_link link;
  
  ASSERT( cvar != NULL );
  ASSERT( sem  != NULL );

  link.synchro = sema;
  link.generic = sem;
  link.pred = NULL;
  SET_DINFO( &link );
  return __kcond_wait( cvar, &link );
}

int kcond_wait_with( kcond_t *cvar, 
					 kcond_synchro_t synchro_type, void *obj,
					 kcond_pred_t *pred )
{
  struct kcond_queue_link link;

  ASSERT( cvar != NULL );
  ASSERT( obj  != NULL );

  link.synchro = synchro_type;
  link.generic = obj;
  link.pred = pred;
  SET_DINFO( &link );
  return __kcond_wait( cvar, &link );
}

int kcond_signal( kcond_t *cvar )
{
  struct kcond_queue_link *queue_head;
  struct kcond_queue_link **queue_prev;

  ASSERT( cvar != NULL );

  SEXP( kcond_stats.signals++ );

  spin_lock( &cvar -> lock );

  queue_prev = &cvar -> queue;
  for( queue_head = cvar -> queue ;
	   queue_head != NULL ; queue_head = queue_head -> next )
	{
	  ASSERT( *queue_prev = queue_head );
	  ASSERT( queue_head -> enqueued );

	  if( ( queue_head -> pred != NULL ) &&
		  ( queue_head -> pred -> pred )( cvar, queue_head -> pred -> data ) )
		{
		  break;
		}
	  else
		{
		  queue_prev = &queue_head -> next;
		}
	}
  if( queue_head == NULL )
	{
	  queue_head = cvar -> queue;
	  queue_prev = &cvar -> queue;
	}
  if( queue_head != NULL )
	{
	  ASSERT( queue_head -> enqueued );
	  ASSERT( queue_prev != NULL );
	  ASSERT( *queue_prev = queue_head );

	  *queue_prev = queue_head -> next;
	  queue_head -> enqueued = 0;
	  up( &queue_head -> wait );
	}
  else
	{
	  SEXP( kcond_stats.lost_signals++ );
	  DEBUG( "Signal with empty queue: %p", cvar );
	}
  spin_unlock( &cvar -> lock );
  return 1;
}

int kcond_broadcast( kcond_t *cvar )
{
  struct kcond_queue_link  *queue_head;
  struct kcond_queue_link **queue_prev;

  ASSERT( cvar != NULL );

  SEXP( kcond_stats.broadcasts++ );

  spin_lock( &cvar -> lock );

  SEXP( ( cvar -> queue == NULL ) ? kcond_stats.lost_broadcasts++ : 0L );

  queue_prev = &cvar -> queue;
  for( queue_head = cvar -> queue ; 
	   queue_head != NULL ; 
	   queue_head = queue_head -> next )
	{
	  ASSERT( queue_head -> enqueued );
	  ASSERT( *queue_prev = queue_head );

	  if( ( queue_head -> pred != NULL ) &&
		  !( queue_head -> pred -> pred )( cvar, queue_head -> pred -> data ) )
		{
		  queue_prev = &queue_head -> next;
		  SEXP( kcond_stats.pred_swtch_skip++ );
		}
	  else
		{
		  queue_head -> enqueued = 0;
		  *queue_prev = queue_head -> next;
		  up( &queue_head -> wait );
		}
	}
  spin_unlock( &cvar -> lock );
  return 1;
}

int kcond_are_waiters( kcond_t *cvar )
{
  ASSERT( cvar != NULL );

  return( cvar -> queue != NULL );
}

void kcond_print( kcond_t *cvar )
{
  if( cvar != NULL )
	{
	  struct kcond_queue_link *item;

	  spin_lock( &cvar -> lock );
	  /* This relies on spinlock_t having .lock field.
		 The same for atomic_t later */
	  PRINT( "%p: lock: %i", cvar, cvar -> lock.lock );
	  DEXP( PRINT( "\t\tcreated: %p by %li", 
				   cvar -> dinfo.return_address, ( long ) cvar -> dinfo.pid ) );
	  PRINT( "waiters:" );
	  for( item = cvar -> queue ; item != NULL ; item = item -> next )
		{
		  PRINT( "\tsemaphore: count: %i sleepers: %i enqueued: %i", 
				 item -> wait.count.counter, item -> wait.sleepers, 
				 item -> enqueued );
		  DEXP( PRINT( "\t\tcreated: %p by %li", 
					   item -> dinfo.return_address, 
					   ( long ) item -> dinfo.pid ) );

		}
	  PRINT( "waiters end." );
	  spin_unlock( &cvar -> lock );
	}
  else
	{
	  PRINT( "<null>" );
	}
}

static int __kcond_wait( kcond_t *cvar, struct kcond_queue_link *link )
{
  int result;

  ASSERT( cvar != NULL );
  ASSERT( link != NULL );

  SEXP( kcond_stats.total_waits++ );
  SEXP( kcond_stats.active_waits++ );

  spin_lock( &cvar -> lock );
  link -> next = cvar -> queue;
  cvar -> queue = link;
  init_MUTEX_LOCKED( &link -> wait );
  link -> enqueued = 1;
  spin_unlock( &cvar -> lock );
  __kcond_synchro_unlock( link -> synchro, link -> generic, cvar );

  result = 0;
  if( link -> enqueued )
	{
	  down_interruptible( &link -> wait );
	  if( signal_pending( current ) )
		{
		  SEXP( kcond_stats.interrupted_waits++ );
		  result = -EINTR;
		}
	}
  else
	{
	  SEXP( kcond_stats.preventive_wakeups++ );
	}
  result = __kcond_synchro_lock( link -> synchro, link -> generic, cvar );

  SEXP( kcond_stats.active_waits-- );
  return result;
}

static int __kcond_synchro_lock( kcond_synchro_t synchro, void *obj,
								 kcond_t *cvar )
{
  ASSERT( obj != NULL );
  
  switch( synchro )
	{
	case sema:
	  {
		return down_interruptible( ( struct semaphore * ) obj );
	  }
	case rw:
	  {
		read_lock( ( rwlock_t * ) obj );
		return 0;
	  }
	case fun:
	  {
		return ( ( ( kcond_synchro_fun_t * ) obj ) -> fun )
		  ( cvar, 
			( ( kcond_synchro_fun_t * ) obj ) -> data );
	  }
	case spin:
	  {
		spin_lock( ( spinlock_t * ) obj );
		return 0;
	  }
	default:
	  {
		ASSERT( 0 );
		return 1;
	  }
	}
}

static int __kcond_synchro_unlock( kcond_synchro_t synchro, void *obj,
								   kcond_t *cvar )
{
  ASSERT( obj != NULL );
  
  switch( synchro )
	{
	case sema:
	  {
		up( ( struct semaphore * ) obj );
		return 0;
	  }
	case rw:
	  {
		read_unlock( ( rwlock_t * ) obj );
		return 0;
	  }
	case fun:
	  {
		return ( ( ( kcond_synchro_fun_t * ) obj ) -> fun )
		  ( cvar, 
			( ( kcond_synchro_fun_t * ) obj ) -> data );
	  }
	case spin:
	  {
		spin_unlock( ( spinlock_t * ) obj );
		return 0;
	  }
	default:
	  {
		ASSERT( 0 );
		return 1;
	  }
	}
}

static int kcond_proc_info( char *page, char **start, 
							off_t off, int count, int *eof, void *data )
{
	int len;

	len = sprintf
	  ( page, 
		"created: %li\n"
		"active: %li\n"
		"total_waits: %li\n"
		"active_waits: %li\n"
		"interrupted_waits: %li\n"
		"signals: %li\n"
		"broadcasts: %li\n"
		"lost_signals: %li\n"
		"lost_broadcasts: %li\n"
		"preventive_wakeups: %li\n",
		"pred_swtch_skip: %li\n",
		kcond_stats.created,
		kcond_stats.active,
		kcond_stats.total_waits,
		kcond_stats.active_waits,
		kcond_stats.interrupted_waits,
		kcond_stats.signals,
		kcond_stats.broadcasts,
		kcond_stats.lost_signals,
		kcond_stats.lost_broadcasts,
		kcond_stats.preventive_wakeups,
		kcond_stats.pred_swtch_skip );

	len -= off;
	if (len < count) 
	  {
		*eof = 1;
		if (len <= 0)
		  {
			return 0;
		  }
	  } 
	else
	  {
		len = count;
	  }
	*start = page + off;
	return len;
}

static int __init kcond_module_init( void )
{
  PRINT( "KCond loading..." );
  kcond_stats.created = 0;
  kcond_stats.active = 0;
  kcond_stats.total_waits = 0;
  kcond_stats.active_waits = 0;
  kcond_stats.interrupted_waits = 0;
  kcond_stats.signals = 0;
  kcond_stats.broadcasts = 0;
  kcond_stats.lost_signals = 0;
  kcond_stats.lost_broadcasts = 0;
  kcond_stats.preventive_wakeups = 0;
  SEXP( create_proc_read_entry
		( "kcond", 0444, &proc_root, kcond_proc_info, NULL ) );
  DEXP( kcond_test() );
  return 0;
}

static void __exit kcond_module_exit( void )
{
  SEXP( remove_proc_entry( "kcond", &proc_root ) );
  PRINT( "KCond unloaded." );
}

module_init( kcond_module_init )
module_exit( kcond_module_exit )

/*
 * $Log$
 */
